/*  
    Sample implementation of proposed C1x mutexes

    Copyright (C) 2010 Just Software Solutions Ltd
    http://www.justsoftwaresolutions.co.uk

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>
*/
#ifndef MUTEX_H
#define MUTEX_H
#include <pthread.h>
#include <linux/futex.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <errno.h>
#include <time.h>


enum mtx_types{
    mtx_plain=0,
    mtx_try=0,
    mtx_timed=0,
    mtx_recursive=1
};

enum thread_return_types{
    thrd_success=0,
    thrd_error,
    thrd_busy,
    thrd_timeout
};

typedef struct xtime
{
    time_t sec;
    long nsec;
} xtime;

typedef struct mutex_t
{
    int flag;
} mutex_t;

typedef struct timed_mutex_t
{
    int flag;
} timed_mutex_t;

typedef struct recursive_mutex_t
{
    mutex_t m;
    unsigned count;
    struct recursive_mutex_t* next;
} recursive_mutex_t;

typedef struct mtx_t
{
    _Bool recursive;
    recursive_mutex_t rmt;
    int flag;
} mtx_t;


inline int mutex_init(mutex_t* m)
{
    m->flag=0;
    return thrd_success;
}

inline int mutex_try_lock(mutex_t* m)
{
    __asm__ __volatile__ ("" :  :  : "memory");
    return __sync_bool_compare_and_swap(&m->flag,0,1)?thrd_success:thrd_busy;
}

inline int mutex_lock(mutex_t* m)
{
    int oldval;
    
    __asm__ __volatile__ ("" :  :  : "memory");
    if(__sync_bool_compare_and_swap(&m->flag,0,1))
    {
        return thrd_success;
    }

    while(oldval=__sync_lock_test_and_set(&m->flag,3))
    {
        if(syscall(SYS_futex, &m->flag,FUTEX_WAIT,oldval,0,0) && 
           (errno!=EWOULDBLOCK) && (errno!=EINTR))
        {
            return thrd_error;
        }
    }
    return thrd_success;
}

inline int mutex_unlock(mutex_t* m)
{
    if(__sync_fetch_and_and(&m->flag,0)>1)
    {
        if(syscall(SYS_futex,&m->flag,FUTEX_WAKE,1,0,0)<0)
        {
            return thrd_error;
        }
    }
    __asm__ __volatile__ ("" :  :  : "memory");
    return thrd_success;
}

inline void mutex_destroy(mutex_t* m)
{}

inline int timed_mutex_init(timed_mutex_t* m)
{
    m->flag=0;
    return thrd_success;
}

inline int timed_mutex_try_lock(timed_mutex_t* m)
{
    __asm__ __volatile__ ("" :  :  : "memory");
    return __sync_bool_compare_and_swap(&m->flag,0,1)?thrd_success:thrd_busy;
}

#define NANOSECONDS_PER_SECOND 1000000000

inline int timed_mutex_try_lock_until(timed_mutex_t* m,const xtime* xt)
{
    int oldval;
    struct timespec target,now,wait_duration;
    
    __asm__ __volatile__ ("" :  :  : "memory");
    if(__sync_bool_compare_and_swap(&m->flag,0,1))
    {
        return thrd_success;
    }

    target.tv_sec=xt->sec;
    target.tv_nsec=xt->nsec;

    while(oldval=__sync_lock_test_and_set(&m->flag,3))
    {
        if(clock_gettime(CLOCK_REALTIME,&now))
        {
            return thrd_error;
        }
        if((now.tv_sec>target.tv_sec) ||
           ((now.tv_sec==target.tv_sec) &&
            (now.tv_nsec>=target.tv_nsec)))
        {
            return thrd_timeout;
        }
        
        wait_duration.tv_sec=target.tv_sec-now.tv_sec;
        wait_duration.tv_nsec=target.tv_nsec-now.tv_nsec;
        if(wait_duration.tv_nsec<0)
        {
            wait_duration.tv_nsec+=NANOSECONDS_PER_SECOND;
            --wait_duration.tv_sec;
        }

        if(syscall(SYS_futex, &m->flag, 
                   FUTEX_WAIT,oldval,&wait_duration,0))
        {
            if(errno==ETIMEDOUT)
            {
                return thrd_timeout;
            }
            else if((errno!=EWOULDBLOCK) && (errno!=EINTR))
            {
                return thrd_error;
            }
        }
        
    }
    return thrd_success;
}

inline int timed_mutex_lock(timed_mutex_t* m)
{
    int oldval;
    
    __asm__ __volatile__ ("" :  :  : "memory");
    if(__sync_bool_compare_and_swap(&m->flag,0,1))
    {
        return thrd_success;
    }

    while(oldval=__sync_lock_test_and_set(&m->flag,3))
    {
        if(syscall(SYS_futex, &m->flag,FUTEX_WAIT,oldval,0,0) && 
           (errno!=EWOULDBLOCK) && (errno!=EINTR))
        {
            return thrd_error;
        }
    }
    return thrd_success;
}

inline int timed_mutex_unlock(timed_mutex_t* m)
{
    if(__sync_fetch_and_and(&m->flag,0)>1)
    {
        if(syscall(SYS_futex,&m->flag,FUTEX_WAKE,1,0,0)<0)
        {
            return thrd_error;
        }
    }
    __asm__ __volatile__ ("" :  :  : "memory");
    return thrd_success;
}

inline void timed_mutex_destroy(timed_mutex_t* m)
{}

inline int recursive_mutex_init(recursive_mutex_t* m)
{
    mutex_init(&m->m);
    m->count=0;
    m->next=0;
}

__thread recursive_mutex_t* owned_mutexes=0;

inline _Bool self_owns_mutex(recursive_mutex_t* m)
{
    recursive_mutex_t* mutex_to_check=owned_mutexes;
    for(;;)
    {
        if(mutex_to_check==m)
        {
            return 1;
        }
        if(!mutex_to_check)
        {
            return 0;
        }
        mutex_to_check=mutex_to_check->next;
    }
}

inline void remove_self_owned_mutex(recursive_mutex_t* m)
{
    recursive_mutex_t* mutex_to_check=owned_mutexes;
    if(mutex_to_check==m)
    {
        owned_mutexes=m->next;
        return;
    }
    
    while(mutex_to_check)
    {
        if(mutex_to_check->next==m)
        {
            mutex_to_check->next=m->next;
            return;
        }
        mutex_to_check=mutex_to_check->next;
    }
}

inline int recursive_mutex_try_lock(recursive_mutex_t* m)
{
    __asm__ __volatile__ ("" :  :  : "memory");
    if(self_owns_mutex(m))
    {
        ++m->count;
        return thrd_success;
    }    
    else if(mutex_try_lock(&m->m)==thrd_success)
    {
        m->next=owned_mutexes;
        owned_mutexes=m;
        m->count=1;
        return thrd_success;
    }
    return thrd_busy;
}

inline int recursive_mutex_lock(recursive_mutex_t* m)
{
    if(recursive_mutex_try_lock(m)==thrd_success)
    {
        return thrd_success;
    }
    else
    {
        int res;
        res=mutex_lock(&m->m);
        if(res!=thrd_success)
        {
            return res;
        }
        m->next=owned_mutexes;
        owned_mutexes=m;
        m->count=1;
        return thrd_success;
    }
}

inline int recursive_mutex_unlock(recursive_mutex_t* m)
{
    if(!--m->count)
    {
        remove_self_owned_mutex(m);
        return mutex_unlock(&m->m);
    }
    return thrd_success;
}

inline void recursive_mutex_destroy(recursive_mutex_t* m)
{}


inline int mtx_init(mtx_t* m, int type)
{
    if(!type&mtx_recursive)
    {
        m->recursive = 0;
        m->flag=0;
        return thrd_success;
    }
    else
    {
        m->recursive=1;
        return recursive_mutex_init(&m->rmt);
    }
    
    return thrd_error;
}

inline int mtx_trylock(mtx_t* m)
{
    int res;
    if(!m->recursive)
    {
        __asm__ __volatile__ ("" :  :  : "memory");
        return __sync_bool_compare_and_swap(&m->flag,0,1)?thrd_success:thrd_busy;
    }
    return recursive_mutex_try_lock(&m->rmt);
}

inline int mtx_lock(mtx_t* m)
{
    if(!m->recursive)
    {
        int oldval;
        
        __asm__ __volatile__ ("" :  :  : "memory");
        if(__sync_bool_compare_and_swap(&m->flag,0,1))
        {
            return thrd_success;
        }
        
        while(oldval=__sync_lock_test_and_set(&m->flag,3))
        {
            if(syscall(SYS_futex, &m->flag,FUTEX_WAIT,oldval,0,0) && 
               (errno!=EWOULDBLOCK) && (errno!=EINTR))
            {
                return thrd_error;
            }
        }
        return thrd_success;
    }
    return recursive_mutex_lock(&m->rmt);
}

inline int mtx_unlock(mtx_t* m)
{
    if(!m->recursive)
    {
        if(__sync_fetch_and_and(&m->flag,0)>1)
        {
            if(syscall(SYS_futex,&m->flag,FUTEX_WAKE,1,0,0)<0)
            {
                return thrd_error;
            }
        }
        __asm__ __volatile__ ("" :  :  : "memory");
        return thrd_success;
    }
    return recursive_mutex_unlock(&m->rmt);
}

inline void mtx_destroy(mtx_t* m)
{
}


#endif
