#include "mutex.h"
#include <time.h>
#include <stdio.h>
#include <assert.h>

#define COUNT 100000000

void time_plain_mutex()
{
    int i;
    clock_t start,mid,mid2,mid3,end;
    mutex_t m;
    recursive_mutex_t rm;
    mtx_t cm;
    pthread_mutex_t pmt;

    mtx_init(&cm,mtx_plain);
    mutex_init(&m);
    recursive_mutex_init(&rm);
    pthread_mutex_init(&pmt,NULL);

    start=clock();

    for(i=0;i<COUNT;++i)
    {
        mutex_lock(&m);
        mutex_unlock(&m);
    }

    mid=clock();
    
    for(i=0;i<COUNT;++i)
    {
        mtx_lock(&cm);
        mtx_unlock(&cm);
    }

    mid2=clock();
    
    for(i=0;i<COUNT;++i)
    {
        recursive_mutex_lock(&rm);
        recursive_mutex_unlock(&rm);
    }

    mid3=clock();
    for(i=0;i<COUNT;++i)
    {
        pthread_mutex_lock(&pmt);
        pthread_mutex_unlock(&pmt);
    }

    end=clock();
    
    printf("Count=%i\n",COUNT);
    printf("Time for mutex_t=%fs\n",((double)(mid-start))/CLOCKS_PER_SEC);
    printf("Time for mtx_t=%fs\n",((double)(mid2-mid))/CLOCKS_PER_SEC);
    printf("Time for recursive_mutex_t=%fs\n",((double)(mid3-mid2))/CLOCKS_PER_SEC);
    printf("Time for pthread_mutex_t=%fs\n",((double)(end-mid3))/CLOCKS_PER_SEC);

    pthread_mutex_destroy(&pmt);
    mtx_destroy(&cm);
    recursive_mutex_destroy(&rm);
    mutex_destroy(&m);
}

void time_recursive_mutex()
{
    int i;
    clock_t start,mid,mid2,end;
    recursive_mutex_t m;
    mtx_t cm;
    pthread_mutexattr_t attr;
    pthread_mutex_t pmt;

    mtx_init(&cm,mtx_plain|mtx_recursive);
    recursive_mutex_init(&m);
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_settype(&attr,PTHREAD_MUTEX_RECURSIVE_NP);
    pthread_mutex_init(&pmt,&attr);

    start=clock();

    for(i=0;i<COUNT;++i)
    {
        recursive_mutex_lock(&m);
        recursive_mutex_lock(&m);
        recursive_mutex_unlock(&m);
        recursive_mutex_unlock(&m);
    }

    mid=clock();
    
    for(i=0;i<COUNT;++i)
    {
        mtx_lock(&cm);
        mtx_lock(&cm);
        mtx_unlock(&cm);
        mtx_unlock(&cm);
    }

    mid2=clock();
    
    for(i=0;i<COUNT;++i)
    {
        pthread_mutex_lock(&pmt);
        pthread_mutex_lock(&pmt);
        pthread_mutex_unlock(&pmt);
        pthread_mutex_unlock(&pmt);
    }

    end=clock();
    
    printf("Count=%i\n",COUNT);
    printf("Time for recursive_mutex_t=%fs\n",((double)(mid-start))/CLOCKS_PER_SEC);
    printf("Time for mtx_t=%fs\n",((double)(mid2-mid))/CLOCKS_PER_SEC);
    printf("Time for pthread_mutex_t=%fs\n",((double)(end-mid2))/CLOCKS_PER_SEC);

    pthread_mutex_destroy(&pmt);
    pthread_mutexattr_destroy(&attr);
    
    mtx_destroy(&cm);
    recursive_mutex_destroy(&m);
}

typedef struct test1_data
{
    int done_flag;
    mutex_t m;
} test1_data;

void* test1_thread(void* p)
{
    int res;
    test1_data* data=(test1_data*)p;
    
    res=mutex_lock(&data->m);
    assert(!res);
    __sync_fetch_and_or(&data->done_flag,1);
    res=mutex_unlock(&data->m);
    assert(!res);
    return 0;
}

void test_plain_mutex_provides_exclusion()
{
    pthread_t t;
    test1_data data;
    struct timespec sleep_time;
    int res;

    data.done_flag=0;
    res=mutex_init(&data.m);
    assert(!res);

    res=mutex_lock(&data.m);
    assert(!res);

    pthread_create(&t,NULL,test1_thread,(void*)&data);

    sleep_time.tv_sec=1;
    sleep_time.tv_nsec=0;
    nanosleep(&sleep_time,NULL);
    
    assert(!__sync_val_compare_and_swap(&data.done_flag,0,0));

    res=mutex_unlock(&data.m);
    assert(!res);

    pthread_join(t,NULL);
    assert(data.done_flag);
}

typedef struct test2_data
{
    int done_flag;
    timed_mutex_t m;
} test2_data;

void* test2_thread(void* p)
{
    int res;
    test2_data* data=(test2_data*)p;
    
    res=timed_mutex_lock(&data->m);
    assert(!res);
    __sync_fetch_and_or(&data->done_flag,1);
    res=timed_mutex_unlock(&data->m);
    assert(!res);
    return 0;
}

void test_timed_mutex_provides_exclusion()
{
    pthread_t t;
    test2_data data;
    struct timespec sleep_time;
    int res;

    data.done_flag=0;
    res=timed_mutex_init(&data.m);
    assert(!res);

    res=timed_mutex_lock(&data.m);
    assert(!res);

    pthread_create(&t,NULL,test2_thread,(void*)&data);

    sleep_time.tv_sec=1;
    sleep_time.tv_nsec=0;
    nanosleep(&sleep_time,NULL);
    
    assert(!__sync_val_compare_and_swap(&data.done_flag,0,0));

    res=timed_mutex_unlock(&data.m);
    assert(!res);

    pthread_join(t,NULL);
    assert(data.done_flag);
}

typedef struct test3_data
{
    int done_flag;
    timed_mutex_t m;
} test3_data;

void* test3_thread(void* p)
{
    int res;
    xtime xt;
    struct timespec ts;
    
    test3_data* data=(test3_data*)p;

    res=clock_gettime(CLOCK_REALTIME,&ts);
    assert(!res);
    
    xt.sec=ts.tv_sec;
    xt.nsec=ts.tv_nsec+NANOSECONDS_PER_SECOND/2;
    if(xt.nsec>NANOSECONDS_PER_SECOND)
    {
        xt.nsec-=NANOSECONDS_PER_SECOND;
        ++xt.sec;
    }

    res=timed_mutex_try_lock_until(&data->m,&xt);
    assert(res==thrd_timeout);
    __sync_fetch_and_or(&data->done_flag,1);
    return 0;
}

void test_timed_mutex_timed_lock_times_out()
{
    pthread_t t;
    test3_data data;
    struct timespec sleep_time;
    int res;

    data.done_flag=0;
    res=timed_mutex_init(&data.m);
    assert(!res);

    res=timed_mutex_lock(&data.m);
    assert(!res);

    pthread_create(&t,NULL,test3_thread,(void*)&data);

    sleep_time.tv_sec=1;
    sleep_time.tv_nsec=0;
    nanosleep(&sleep_time,NULL);
    
    assert(__sync_val_compare_and_swap(&data.done_flag,0,0));

    res=timed_mutex_unlock(&data.m);
    assert(!res);

    pthread_join(t,NULL);
    assert(data.done_flag);
}

typedef struct test4_data
{
    int done_flag;
    timed_mutex_t m;
} test4_data;

void* test4_thread(void* p)
{
    int res;
    xtime xt;
    struct timespec ts;
    
    test4_data* data=(test4_data*)p;
    
    res=clock_gettime(CLOCK_REALTIME,&ts);
    assert(!res);
    
    xt.sec=ts.tv_sec+2;
    xt.nsec=ts.tv_nsec;

    res=timed_mutex_try_lock_until(&data->m,&xt);
    assert(res==thrd_success);
    __sync_fetch_and_or(&data->done_flag,1);
    return 0;
}

void test_timed_mutex_timed_lock_takes_lock_if_in_time()
{
    pthread_t t;
    test4_data data;
    struct timespec sleep_time;
    int res;

    data.done_flag=0;
    res=timed_mutex_init(&data.m);
    assert(!res);

    res=timed_mutex_lock(&data.m);
    assert(!res);

    pthread_create(&t,NULL,test4_thread,(void*)&data);

    sleep_time.tv_sec=0;
    sleep_time.tv_nsec=NANOSECONDS_PER_SECOND/2;
    nanosleep(&sleep_time,NULL);

    res=timed_mutex_unlock(&data.m);
    assert(!res);

    pthread_join(t,NULL);
    assert(data.done_flag);
}


int main()
{
    test_plain_mutex_provides_exclusion();
    test_timed_mutex_provides_exclusion();
    test_timed_mutex_timed_lock_times_out();
    test_timed_mutex_timed_lock_takes_lock_if_in_time();
    
    time_plain_mutex();
    time_recursive_mutex();
    return 0;
}


